;==================================================================
;
;                          resprop.scm
;
;                   Gandalf version c-2.1
;
;    
;       Propositional prover & finite satisfiability checker.
;
;
;====================================================================
;
; Tasks:
; 1. Satisfiability proving on finite domains: find a finite domain where all sat.
;    Funs with fixed args are fully interpreted to selected constants.
;     
; 2. Unsatisfiability proving by instantiating on one domain (initially
;    function-free) and finding a contradiction for that instantiation.
; 
; 3. Unsatisfiability proving by instantiating on one domain with functions 
;    where functions are interpreted either to new constants or according
;    to the equality units in the formula and finding a contradiction
;    for that instantiation.     
;        
; 4. Highly incomplete unsatisfiability proving by instantiating by giving all vars
;    one concrete constant value and finding a contradiction
;    for that instantiation.   
;
;
;
; Algorithm:
;
; run over all existing clauses, starting with shorter and smaller, for each clause CL do:
;    for each var X in CL do:
;      for all the values V in *prop-constants* do:
;         give X value V
;         for each fully instantiated clause CLI do:
;             compute the values of funs by using mapping *prop-fun-mappings*,
;               generate clause CLIV
;             run over all the literals and find their value using mapping 
;               *prop-lit-mappings* and checking: 
;                       if value is #t, discard full clause
;                       if value is #f, discard this particular literal
;                          if all lits discarded, contradiction found                       
;                       else insert value nr either pos or neg
;
; generate a list of vectors of numbers, each having 
; a first nr indicating an id, sec len, rest lits.
; generate a vector of lits, each indicating a value of a lit #t, #f or null.
; start binary ordered resolution, preferring smaller clauses. If horn, use unit res.
; If a unit lit is generated, store the value in the value vector.
; generate a subsumption tree
;
;
; Finite domain construction:
;
;  Cycle over domain sizes 1, 2, 3, etc:
;     For each constant symb and fun symb build all mappings for selected domain
;       Select one mapping for each const and fun, run over all such selections:
;           check satisfiability of the mapping and domain. 
;           if satisfibale: we are finished, sat proved
;           if not satisfiable: continue.


;----------------------------------------------------------------------
;
;                          compiler defs
;
;----------------------------------------------------------------------

;(define compile-inline 
;  '(litinf-left-para? litinf-right-para? litinf-parainto? 
;    pos-eq-pred? pos-eq-fun? litinf-binrespara?		      
;    eq-lit-arg1 eq-lit-arg2 eq-lit-type may-be-demodulated?
;    litinf-use make-binres-history))

;----------------------------------------------------------------------
; 
;                 global data
;
;----------------------------------------------------------------------



;Original pred-level data lists:

(define *prop-orig-constants* '()) ; list of constants
(define *prop-orig-funs* '())  ; alist with funs and arities: ((f . 2)(g . 1))
(define *prop-orig-preds* '()) ; alist with preds and arities: ((p . 2)(r . 0))

(define *prop-flat-eq-symb* '()) ; enumeration num corresponding to equality

;Prop-level main lists:

(define *prop-constants* '()) ; selection from prop-orig-constants

(define *prop-constants-mapping* '()) ; alist mapping from orig constants to new ones

(define *prop-fun-mappings* '()) ; for all funs (f,g) and arg trees 
				 ; consisting of prop-constants:
	;   ( (f (1 (1 . val)
	;           (2 . val)
	;           (3 . val) )
	;        (2 (1 . val)
	;           (2 . val)
	;           (3 . val) ) 
	;        (3 (1 . val)
	;           (2 . val)
	;           (3 . val) ))                         
	;     (g  ...) )

(define *prop-lit-mappings* '()) ; analogous to *prop-fun-mappings*, 
				 ; val may be nr, #t or #f

(define *prop-atom-maxnr* 10000) ; enumerates all prop atoms

(define *prop-interp-value-pairs* '()) ; a list of pairs where each cdr should
				       ; replaced with a value from domain	

(define *prop-interp-termlist* '()) ; a list of terms which will be interpreted

(define *tmp-varlist* '())

(define *prop-orig-constants-enum* '())
(define *prop-orig-funs-enum* '())
(define *prop-orig-preds-enum* '())
(define *prop-last-enum-value* 0)    

(define *propclauses* '())
(define *propclause-tree* '())
(define *prop-time-ran-out* #f)

;----------------------------------------------------------------------
; 
;       top level of satisfiability checking with model building
;
;----------------------------------------------------------------------

; prop-build-simple-model returns #t iff satisfiable, #f if building
; failed for any reason 

(define (prop-build-simple-model predclauses)
  (let ((stopflag #f)   
        (domainsize 0)  
        (tmpnr 0)      
        (res #f) )        
    (set! *prop-time-ran-out* #f)  
    (set! *given-checkpoint-keptnrs* '())    
    (set! *slow-start-time*  (get-check-time 0))
    (store-time-checkpoint! 0)        
    (extract-pred-data-to-prop predclauses)
    
    ;(for-each (lambda(x)(newline)(display x)) predclauses)
    ;(display "prop-orig-funs: ")(display *prop-orig-funs*)(newline)
    
    (cond            
      ((null? *prop-orig-funs*) 
      ; no function symbols: interpret all constants to themselves
            (set! domainsize (length *prop-orig-constants*))
            (if (zero? domainsize) (set! domainsize 1))
            (if (> *display-info-level* 38) 
               (begin ; (newline)
	              (display "no function symbols, starting search for domain size ")
	              (display domainsize) 
		      (newline) ))
                ; last arg #t iff const are interp
           (set! *prop-interp-value-pairs* '())               
           (make-prop-interpretation-lists domainsize #t #f)
           ;(display *prop-interp-value-pairs* )           
           (create-prop-var-vectors *prop-atom-maxnr*) 	           
           ;(display *prop-interp-value-pairs* )
           ;(abort)
           ; now try to solve                                         	      
           (set! res (check-prop-one-interpretation predclauses domainsize))
           (if *prop-time-ran-out* 
    	       (begin
	               (if (> *display-info-level* 38) 
           	           (begin (newline)
	                          (display "allocated model search time has run out")	                      
		                  (newline) ))
	                #f)
               res))
      ; function symbols present: full domain size iteration  
     (else   
    	(do ((domainsize 1 (+ 1 domainsize)))
            ((or stopflag
                 res
                 (> (- (get-check-time 1) *slow-start-time*)
	            *one-pass-time*))
	        ;(if (and (not stopflag)(not res))
	        ;    (store-time-checkpoint! 1) )   
	        (if *prop-time-ran-out* 
    	            (begin
	               (if (> *display-info-level* 38) 
           	           (begin (newline)
	                          (display "allocated model search time has run out")	                      
		                  (newline) ))
	                #f)
                    res))
           (if (> *display-info-level* 38) 
               (begin ; (newline)
	              (display "starting search for domain size ")
	              (display domainsize) 
		      (newline) ))
        
           (make-prop-interpretation-lists domainsize #t #f) ; 
           (create-prop-var-vectors *prop-atom-maxnr*)
       
           (if *prop-debug*
            (begin
             (display "*prop-interp-value-pairs*: ")(newline)
             (display *prop-interp-value-pairs*)(newline)
             (display "*prop-constants-mapping*: ")(newline)
             (display *prop-constants-mapping*)(newline) 
             (display "*prop-fun-mappings*: ") (newline)
             (display *prop-fun-mappings*)(newline)
             (display "*prop-lit-mappings*: ") (newline)
             (display *prop-lit-mappings*)(newline)    
             (display "*prop-atom-maxnr*: ")(newline)
             (display *prop-atom-maxnr*)(newline) ))       
       
           (set! res (check-prop-interpretations predclauses domainsize 0)) )))))
       
            ; (if (> domainsize 2) (set! stopflag #t)) ))))  
 

(define (prop-build-flatten-model predclauses)
  (let ((stopflag #f)   
        (domainsize 0)  
        (flatclauses '())
        (tmpnr 0)      
        (res #f) )        
    ; (set! *prop-atom-maxnr* *prop-last-enum-value*)    
    (set! *prop-time-ran-out* #f)  
    (set! *given-checkpoint-keptnrs* '())    
    (set! *slow-start-time*  (get-check-time 0))
    (store-time-checkpoint! 0)        
    (extract-pred-data-to-prop predclauses)
    (set! flatclauses (flatten-clause-set predclauses))
    
        ; full domain size iteration  
    	(do ((domainsize 1 (+ 1 domainsize)))
            ((or stopflag
                 res
                 (> (- (get-check-time 1) *slow-start-time*)
	            *one-pass-time*))
	        ;(if (and (not stopflag)(not res))
	        ;    (store-time-checkpoint! 1) )   
	        (if *prop-time-ran-out* 
    	            (begin
	               (if (> *display-info-level* 38) 
           	           (begin (newline)
	                          (display "allocated model search time has run out")	                      
		                  (newline) ))
	                #f)
                    res))
           (if (> *display-info-level* 38) 
               (begin ; (newline)
	              (display "starting search for domain size ")
	              (display domainsize) 
		      (newline) ))
        
           (make-prop-interpretation-lists domainsize #f #t) ;
           
           (create-prop-var-vectors *prop-atom-maxnr*)
       
           (if *prop-debug*
            (begin
             (display "*prop-interp-value-pairs*: ")(newline)
             (display *prop-interp-value-pairs*)(newline)
             (display "*prop-constants-mapping*: ")(newline)
             (display *prop-constants-mapping*)(newline) 
             (display "*prop-fun-mappings*: ") (newline)
             (display *prop-fun-mappings*)(newline)
             (display "*prop-lit-mappings*: ") (newline)
             (display *prop-lit-mappings*)(newline)    
             (display "*prop-atom-maxnr*: ")(newline)
             (display *prop-atom-maxnr*)(newline) ))       
       
           (set! res (check-prop-flatten-interpretations
                         predclauses flatclauses domainsize)) )))
       
            ; (if (> domainsize 2) (set! stopflag #t)) ))))  
 


; prop-build-random-model returns #t iff satisfiable, #f if building
; failed for any reason 

(define (prop-build-random-model predclauses)
  (let ((stopflag #f)   
        (domainsize 0)  
        (randomsize 20)
        (tmpnr 0)              
        (res #f) )        
    (set! *prop-time-ran-out* #f)  
    (set! *given-checkpoint-keptnrs* '())    
    (set! *slow-start-time*  (get-check-time 0))
    (store-time-checkpoint! 0)        
    (extract-pred-data-to-prop predclauses)
  
    (do ((domainsize 1 (+ 1 domainsize)))
        ((or stopflag
             (> domainsize 30)
             res
                 (> (- (get-check-time 1) *slow-start-time*)
	            *one-pass-time*))
	        ;(if (and (not stopflag)(not res))
	        ;    (store-time-checkpoint! 1) )   
	        (if *prop-time-ran-out* 
    	            (begin
	               (if (> *display-info-level* 38) 
           	           (begin (newline)
	                          (display "allocated model search time has run out")	                      
		                  (newline) ))
	                #f)
                    res))
           (if (> *display-info-level* 38) 
               (begin ; (newline)
	              (display "starting search for domain size ")
	              (display domainsize) 
		      (newline) ))
        
           (make-prop-interpretation-lists domainsize #t #f) ; 
           (create-prop-var-vectors *prop-atom-maxnr*)
       
           (if *prop-debug*
            (begin
             (display "*prop-interp-value-pairs*: ")(newline)
             (display *prop-interp-value-pairs*)(newline)
             (display "*prop-constants-mapping*: ")(newline)
             (display *prop-constants-mapping*)(newline) 
             (display "*prop-fun-mappings*: ") (newline)
             (display *prop-fun-mappings*)(newline)
             (display "*prop-lit-mappings*: ") (newline)
             (display *prop-lit-mappings*)(newline)    
             (display "*prop-atom-maxnr*: ")(newline)
             (display *prop-atom-maxnr*)(newline) ))       
       
           (set! res (check-prop-interpretations predclauses domainsize randomsize)) )))
       
            ; (if (> domainsize 2) (set! stopflag #t)) ))))  
 


;----------------------------------------------------------------------
; 
;                 pred-level data extraction, initialisation
;
;----------------------------------------------------------------------

; takes a full fully-built clause list and extracts all constant, fun and pred symbs
; assigns:
; *prop-orig-constants*
; *prop-orig-funs*
; *prop-orig-preds*

(define (extract-pred-data-to-prop predclauselist)
  (let () 
     (set! *prop-orig-constants* '())
     (set! *prop-orig-funs* '())     
     (set! *prop-orig-preds* '())         
     (set! *prop-last-enum-value* 0)         
     
     ; build lists of original constants, funs and preds
     
     (for-each 
        (lambda(fullcl)
          (if (and (pair? fullcl)(pair? (cdr fullcl)))
              (for-each 
                (lambda(lit)
                  (extract-pred-data-to-prop-aux-lit lit) )
                (cdr fullcl) )))                  
        predclauselist)
        
     (set! *prop-orig-constants-enum* 
           (map (lambda(x)
                  (set! *prop-last-enum-value* (+ 1 *prop-last-enum-value*))
                  (cons x *prop-last-enum-value*) )
                *prop-orig-constants*))
     (set! *prop-orig-funs-enum* 
           (map (lambda(x)
                  (set! *prop-last-enum-value* (+ 1 *prop-last-enum-value*))
                  (cons (car x) *prop-last-enum-value*) )
                *prop-orig-funs*))            
     (set! *prop-orig-preds-enum* 
           (map (lambda(x)    
                  (set! *prop-last-enum-value* (+ 1 *prop-last-enum-value*))
                  (if (pos-eq-pred? (car x)) 
                      (set! *prop-flat-eq-symb* *prop-last-enum-value*) )
                  (cons (car x) *prop-last-enum-value*) )
                *prop-orig-preds*))    

     (if *prop-debug*
          (begin
     		(display *prop-orig-constants*)
     		(display *prop-orig-funs*)
     		(display *prop-orig-preds*)
     
     		(display *prop-orig-constants-enum*)
     		(display *prop-orig-funs-enum*)
     		(display *prop-orig-preds-enum*) ))))
    


     
(define (extract-pred-data-to-prop-aux-lit lit)
  (let ((symb '())
        (argpart '()) ) 
     (set! symb (pure-unsigned-fun (car lit))) 
     (if (not (assq symb *prop-orig-preds*))
         (set! *prop-orig-preds* 
               (cons (cons symb (length (cdr lit))) *prop-orig-preds*) ))
     (do ((argpart (cdr lit) (cdr argpart)))
         ((null? argpart))
        (extract-pred-data-to-prop-aux-fun (car argpart)) )))
        

(define (extract-pred-data-to-prop-aux-fun term)
  (let ((symb '())
        (argpart '()) ) 
     (cond ((pair? term)  
     	      (set! symb (pure-fun (car term))) 
     	      (if (not (assq symb *prop-orig-funs*))
                  (set! *prop-orig-funs* 
                        (cons (cons symb (length (cdr term))) *prop-orig-funs*) ))
     	      (do ((argpart (cdr term) (cdr argpart)))
         	  ((null? argpart))
        	(extract-pred-data-to-prop-aux-fun (car argpart)) ))
           ((const? term)
              (set! symb term)
              (if (not (memq symb *prop-orig-constants*))
                  (set! *prop-orig-constants* (cons symb *prop-orig-constants*)) )))))
                      	         
     

;----------------------------------------------------------------------
; 
;                 interpretation lists construction 
;
;----------------------------------------------------------------------
;
;
;  const a, b, c to 1,2:
;   a=1
;   b=1
;   c=2
;  
;  fun f with arity 2 to 1,2:
;   f(1,1)=1
;   f(1,2)=1
;   f(2,1)=2
;   f(2,2)=2

; make-prop-interpretation-lists creates interpretation lists to be
; assigned and backtraced over later. If interpconstants is #f, constant
; symbols will not be put there.


(define (make-prop-interpretation-lists domainsize interpconstants flattenflag)    
  (let ((tmp '())  
        (modified '()) ) 
     (set! *prop-interp-value-pairs* '())
     (set! *prop-interp-termlist* '())
     (set! *prop-constants-mapping* '()) 
     (set! *prop-fun-mappings* '()) 
     (set! *prop-lit-mappings* '())     
     (set! *prop-atom-maxnr* 0)
     (if flattenflag
       ; interpret both preds, funs and constants as preds (plus one arg ...)
       (begin
         (set! modified 
                 (append 
       		    (map (lambda(x)
       		           (set! tmp (assq (car x) *prop-orig-preds-enum*))
       		           (cons (cdr tmp) (cdr x)) )
       		         *prop-orig-preds*)
       		   (append
       		        (map (lambda(x)
       		               (set! tmp (assq (car x) *prop-orig-funs-enum*))
       		               (cons (cdr tmp)(+ 1 (cdr x))))
       		             *prop-orig-funs*)
       		        (map (lambda(x)
       		               (set! tmp (assq x *prop-orig-constants-enum*))
       		               (cons (cdr tmp) 1))
       		             *prop-orig-constants*))))  
       		              
         ; (newline)(display "modified prop mapping input: ")
         ; (newline)(display modified)
       
         (set! *prop-lit-mappings*
           (map (lambda(funinfo)  	        
                  (cons (car funinfo)
                        (make-prop-interpretation-lists-aux 
                           (cdr funinfo) (list (car funinfo)) 
                           domainsize #f flattenflag) ))
       		modified)))	           	 	       
       ; normal interpretationmark
       (begin
       	(set! *prop-lit-mappings*
           (map (lambda(funinfo)  	        
                  (cons (car funinfo)
                        (make-prop-interpretation-lists-aux 
                           (cdr funinfo) (list (car funinfo)) 
                           domainsize #f flattenflag) ))
       		*prop-orig-preds*))     	
        (set! *prop-fun-mappings*
           (map (lambda(funinfo)  	        
                  (cons (car funinfo)
                        (make-prop-interpretation-lists-aux 
                          (cdr funinfo) (list (car funinfo))
                          domainsize #t flattenflag) ))
       		*prop-orig-funs*))
        (set! *prop-constants-mapping*
     	   (map (lambda(const)
     	          (set! tmp (cons const '())) 
     	          (if interpconstants
     	              (begin 
     	                 (set! *prop-interp-termlist*     	          
     	          	      (cons (cons const '()) *prop-interp-termlist*) )
     	          	 (set! *prop-interp-value-pairs*     	          
     	          	      (cons tmp *prop-interp-value-pairs*) )))
     	          tmp)	
     	        *prop-orig-constants*))))))		

       		            
(define (make-prop-interpretation-lists-aux 
             argsleft term domainsize addinterpflag flattenflag) 

  ; (newline)(display "make-prop-interpretation-lists-aux called: ")
  ; (newline)(display argsleft)(newline)
  ; (display term)(newline)


  (let ((nr 0)
        (tmp '())
        (res '()) )
    (do ((nr 0 (+ 1 nr)))
        ((>= nr domainsize) res)
      (if (<= argsleft 1)    
          (if addinterpflag 
              (begin
                (set! tmp (cons nr '()))
                (set! *prop-interp-value-pairs*
     	              (cons tmp *prop-interp-value-pairs*) )     	           
     	        (set! *prop-interp-termlist*      
     	              (cons (reverse (cons nr term)) *prop-interp-termlist*) )
     	        (set! res (cons tmp res)) )
     	      (begin
     	        (set! *prop-atom-maxnr* (+ 1 *prop-atom-maxnr*)) 
     	        (set! tmp (cons nr *prop-atom-maxnr*))     	       
     	        (set! res (cons tmp res)) ))     
          (begin (set! tmp (cons nr (make-prop-interpretation-lists-aux 
                                        (- argsleft 1) 
                                        (cons nr term)
                                        domainsize addinterpflag flattenflag)))
                 (set! res (cons tmp res)) )))))
                   		            
    
         
;----------------------------------------------------------------------
; 
;       pred-level formula saturation with domain values, 
;       prop form construction
;
;----------------------------------------------------------------------
    
; check-prop-one-interpretations  returns #t iff it finds a model for 
; a given domain size and given interpretation
; ie iff the predclauses are found satisfiable
; a value #f means a failure to find model
;
; it DOES NOT generate new interpretations of fun symbs
    
    
(define (check-prop-one-interpretation predclauses domainsize)  
  (let ((valuations '())
        (selectedvals '()) 
        (pre-evalres #f) 
        (evalres #f) 
        (allchecked #f)
        (tmpnr -1)
        (maxval (- domainsize 1)) )        
       (set! *propclause-tree* '())   
       ; create a simple valuation for all the constants
       (set! valuations (map (lambda(x)
                                  (set! tmpnr (+ 1 tmpnr)) 
                                  tmpnr) 
                             *prop-interp-value-pairs*))
       ; insert the newly created valuations                      
       (insert-valuations! *prop-interp-value-pairs* valuations)  
       
       (if *prop-debug*
          (begin
            (newline)
            (display "valuations: ")(display valuations) ))   
          
       (if (> *display-info-level* 50)
         (begin
           (for-each (lambda(x y)
                       (display x) (display " = ") (display y) (newline) )                
                     *prop-interp-termlist*
                     valuations)
           (newline) ))        
   
   
       (if (> *display-info-level* 70) 
           (begin (newline)
	          (display "starting to check satisfiability for symbol interp:")
	          (newline)
	          (display *prop-interp-value-pairs*)
		  (newline) ))

       ; clear the vector used for storing values during clause conversion to prop
       
       (vector-fill! *propvar-enqued-value* '())
                
       ; make-predclauses-to-propclauses gives #t if might be sat, 
        ; #f iff certainly not satisfiable
       
       (set! pre-evalres (make-predclauses-to-propclauses predclauses domainsize #f))
       
       (if (> *display-info-level* 70) 
           (begin (newline)
	          (if pre-evalres 
	             (display "conversion to prop did not find unsatisfiability")
	             (display "conversion to prop found unsatisfiability") )
		  (newline) ))
       
       (if *prop-debug*
          (begin
             (newline)
             (display "pre-evalres: ")(display pre-evalres) ))
       
       ; solve-prop-clauses gives #t if satisfiable, #f if not satisfiable
       
       (if pre-evalres 
           (set! evalres (solve-prop-clauses *propclauses* *prop-atom-maxnr*)) )           
    
       (if (> *display-info-level* 70) 
           (begin (newline)
	          (if evalres
	             (display "full prop evaluation found prop satisfiable")
	             (display "full prop evaluation found prop unsatisfiable") )
		  (newline) ))
       
       (if *prop-debug*
          (begin (display "funsymb interp vals: ")(display valuations)(newline)) )
          
      (if *prop-debug*
          (begin (newline)(display "evalres: ")(display evalres)(newline)) )
         
      evalres))
        
   
   

       
; check-prop-flatten-interpretations  returns #t iff it finds a model for 
; a given domain size
; ie iff the predclauses are found satisfiable
; a value #f means a failure to find model
;
; NB! no function or constant symbols are present!
   
    
(define (check-prop-flatten-interpretations predclauses flatclauses domainsize)  
  (let ((valuations '())
        (selectedvals '()) 
        (pre-evalres #f) 
        (evalres #f) 
        (extraclauses '())
        (allchecked #f)
        (tmpnr -1)
        (maxval (- domainsize 1)) )              
       (set! *propclause-tree* '())  
       ; clear the vector used for storing values during clause conversion to prop
       
       (vector-fill! *propvar-enqued-value* '())
                
       ; make-predclauses-to-propclauses gives #t if might be sat, 
       ; #f iff certainly not satisfiable
       
       (set! extraclauses (make-extra-flatten-propclauses domainsize))
       (if #f
           (begin
            (newline) 
            (for-each (lambda(x)(display x)(newline)) extraclauses) 
            (newline) ))
           
       (set! flatclauses 
            (append extraclauses flatclauses) )
       
       (set! pre-evalres (make-predclauses-to-propclauses flatclauses domainsize #t))       
       
       
       (if (> *display-info-level* 70) 
           (begin (newline)
	          (if pre-evalres 
	             (display "conversion to prop did not find unsatisfiability")
	             (display "conversion to prop found unsatisfiability") )
		  (newline) ))
       
       (if *prop-debug*
          (begin
             (newline)
             (display "pre-evalres: ")(display pre-evalres) ))
       
       ; solve-prop-clauses gives #t if satisfiable, #f if not satisfiable
       
       (if pre-evalres 
           (set! evalres (solve-prop-clauses *propclauses* *prop-atom-maxnr*)) )           
    
       (if (> *display-info-level* 70) 
           (begin (newline)
	          (if evalres
	             (display "full prop evaluation found prop satisfiable")
	             (display "full prop evaluation found prop unsatisfiable") )
		  (newline) ))
                  
      (if *prop-debug*
          (begin (newline)(display "evalres: ")(display evalres)(newline)) )
         
      evalres))
        
          
       
; check-prop-interpretations  returns #t iff it finds a model for 
; a given domain size
; ie iff the predclauses are found satisfiable
; a value #f means a failure to find model
;
; it generates and tests all interpretations of fun symbs

  
(define (check-prop-interpretations predclauses domainsize randomsize)
  (let ((valuations '())
        (selectedvals '()) 
        (pre-evalres #f) 
        (evalres #f)
        (count 0) 
        (allchecked #f)
        (maxval (- domainsize 1)) )        
     (set! *propclause-tree* '())     
     (set! valuations (map (lambda(x) 0) *prop-interp-value-pairs*))
     (do ((count 0 (+ 1 count)))
         ((or evalres 
              allchecked
              (and (> randomsize 0) (> count randomsize))
              (> (- (get-check-time 1) *slow-start-time*)
	         *one-pass-time*))
	    ;(if (and (not evalres)(not allchecked))
	    ;    (store-time-checkpoint! 1) )  
            
            (if (and (not evalres)(not allchecked))
                (set! *prop-time-ran-out* #t) )
            
            (if *prop-debug*
               (begin (newline)(display "evalres: ")(display evalres)(newline)))
          
            evalres)              
       (insert-valuations! *prop-interp-value-pairs* valuations)
       
       (if *prop-debug*
          (begin
            (newline)
            (display "valuations: ")(display valuations) ))   
          
       (if (> *display-info-level* 50)
         (begin
           (for-each (lambda(x y)
                       (display x) (display " = ") (display y) (newline) )                
                     *prop-interp-termlist*
                     valuations)
           (newline) ))        
   
   
       (if (> *display-info-level* 70) 
           (begin (newline)
	          (display "starting to check satisfiability for symbol interp:")
	          (newline)
	          (display *prop-interp-value-pairs*)
		  (newline) ))

       ; clear the vector used for storing values during clause conversion to prop
       
       (vector-fill! *propvar-enqued-value* '())
                
       ; make-predclauses-to-propclauses gives #t if might be sat, 
       ; #f iff certainly not satisfiable
       
       (set! pre-evalres (make-predclauses-to-propclauses predclauses domainsize #f))
       
       (if (> *display-info-level* 70) 
           (begin (newline)
	          (if pre-evalres 
	             (display "conversion to prop did not find unsatisfiability")
	             (display "conversion to prop found unsatisfiability") )
		  (newline) ))
       
       (if *prop-debug*
          (begin
             (newline)
             (display "pre-evalres: ")(display pre-evalres) ))
       
       ; solve-prop-clauses gives #t if satisfiable, #f if not satisfiable
       
       (if pre-evalres 
           (set! evalres (solve-prop-clauses *propclauses* *prop-atom-maxnr*)) )           
    
       (if (> *display-info-level* 70) 
           (begin (newline)
	          (if evalres
	             (display "full prop evaluation found prop satisfiable")
	             (display "full prop evaluation found prop unsatisfiable") )
		  (newline) ))
    
       (if *prop-debug*
          (begin (display "funsymb interp vals: ")(display valuations)(newline)) )
        
       (if (> randomsize 0)
           (valuation-list-nextrandomvalues valuations valuations domainsize)
           (set! allchecked 
                (or (null? valuations)
                    (valuation-list-nextvalues valuations valuations domainsize) ))))))
 
          
(define (valuation-list-nextvalues valuationlist origvaluations domainsize)   

    ;(begin (display "in nextvalues: ")
    ;       (display valuationlist)(display " ")
    ;       (display origvaluations)(display " ")
    ;       (display domainsize)(newline) ) 
   
    (if (and (pair? valuationlist)
             (or (and (pair? origvaluations)
                      (> (car valuationlist)
                            (max-of-two (max-of-list (cdr valuationlist))
                                        (max-of-corr-fun-args 
                                          valuationlist origvaluations) )))             
                 (>= (car valuationlist) (- domainsize 1)) ))
        (begin        
              (if (null? (cdr valuationlist))
                   #t 
                   (begin
                     (set-car! valuationlist 0)
                     (valuation-list-nextvalues
                         (cdr valuationlist) origvaluations domainsize) )))                     
        (begin (set-car! valuationlist (+ 1 (car valuationlist)))
               #f)))
                 
                 
(define (max-of-two one two)
  (if (> one two) one two) )                 

(define (max-of-list lst)
  (let ((res 0))
     (do ((part lst (cdr part))) 
         ((null? part) res)
       (if (> (car lst) res) (set! res (car lst))) )))  
 
(define (max-of-corr-fun-args valuationlist origvaluationlist)
   (let ((term (get-valuationlistpos-term valuationlist origvaluationlist)))  
      ;(display term)
      (if (pair? term) 
          (max-of-list (cdr term))
          term)))

(define (get-valuationlistpos-term valuationlist origvaluationlist)
    ;(display valuationlist)(display origvaluationlist)(display *prop-interp-termlist*)
    (let ((res '()))
       (do ((part1 origvaluationlist (cdr part1))
            (part2 *prop-interp-termlist* (cdr part2)) )            
           ((or (null? part1) (eq? part1 valuationlist)) 
               (if (null? part1) 
                   999999
                   (car part2) )))))
                    
          
(define (insert-valuations! value-pairs valuations)
  (for-each (lambda(x y)(set-cdr! x y))
             value-pairs
             valuations))



(define (valuation-list-nextrandomvalues valuationlist origvaluations domainsize)   
  
    ;(begin (display "in nextrandomvalues: ")
    ;       (display valuationlist)(display " ")(display domainsize)(newline)) 
  (let ((tmp 0)) 
    (do ((part valuationlist (cdr part)))
        ((null? part))
       (set! tmp (random-random domainsize))
       (if (or (not (integer? tmp)) (< tmp 0) (>= tmp domainsize)) 
           (begin
              (display "bad random: ")(display tmp)(newline)
              (set! tmp 0) ))
       (set-car! part tmp) )))
    
             
             
; sets the global value *propclauses*

; make-predclauses-to-propclauses returns #f iff found to be unsatisfiable,
; else #t

(define (make-predclauses-to-propclauses predclauses domainsize flatflag)
  (let ((tmp '())
  	(unsatisfiable #f)  
        (predclausespart '()) )
    (set! *propclauses* '())  
    (set! *propclause-tree* '())    
    (if (not flatflag)        
        (set! predclauses (sort-predclauses-for-propconversion predclauses)) )
    
    (if *prop-debug*
          (begin
     		(display "predclauses before conversion: ")(newline)
    		(newline)(for-each (lambda(x)(newline)(display x)) predclauses)(newline)
    		(display "now convert: ")(newline) ))
        
    (do ((predclausespart predclauses (cdr predclausespart)))
        ((or (null? predclausespart) unsatisfiable)
           
           (if *prop-debug*
          	(begin
           	 (newline)(display "unsatisfiable:")(display unsatisfiable)(newline)))
           
           (not unsatisfiable))
           
       (if *prop-debug*
          	(begin (display (car predclausespart))(newline)) )    
           
      (if (and (pair? (car predclausespart)) 
               (pair? (cdar predclausespart)))
          (begin
            (set! tmp 
               (make-predclause-to-propclauses 
                   (car predclausespart) domainsize flatflag))
                   
            (if *prop-debug*
                (begin (newline)(display "tmp: ")(display tmp)(newline)) )       
                   
            (if tmp (set! unsatisfiable #t)) )))))
               

(define (sort-predclauses-for-propconversion predclauses)
  (let ((cl '())
        (res1 '())
        (res2 '())
        (res3 '()) 
        (res '()) )
    (do ((part predclauses (cdr part)))
        ((null? part) 
          (append (append res1 res2) res3))                 
      (set! cl (car part))                  
      (cond 
         ((not (pair? cl)))
         ((not (pair? (cdr cl))))
         ((and (null? (cddr cl))
               (fun-may-cont-vars? (car (cadr cl))) )
            (set! res2 (cons cl res2)) )   
         ((null? (cddr cl))
            (set! res1 (cons cl res1)) )
         (else      
            (set! res3 (cons cl res3)) ))))) 


;(define (sort-clause clause pair-flag)
;;
;(define (sort-clause-aux cl len pair-flag)
;  (cond ((< len 2) cl)
;	((eq? len 2)
;	   (if (if pair-flag
;                   (>-fun-ordering? (cadar (cdr cl)) (cadar cl))
;                   (>-fun-ordering? (caadr cl) (caar cl)) )
;	       (list (cadr cl) (car cl))
;	       cl))
;	(else 
;	  (let* ((l1 (quotient len 2))
;		 (l2 (- len l1)) 
;		 (s (split-list cl l1)) )
;	    (sort-merge-clauses 
;	       (sort-clause-aux (car s) l1 pair-flag)
;	       (sort-clause-aux (cdr s) l2 pair-flag)
;               pair-flag)))))
;	
;(define (sort-merge-clauses cl1 cl2 pair-flag)
;  (cond ((null? cl1) cl2)
;	((null? cl2) cl1)
;        ((and pair-flag (>-fun-ordering? (cadar cl1) (cadar cl2)))
;           (cons (car cl1)
;                 (sort-merge-clauses (cdr cl1) cl2 #t) ))
;	((and (not pair-flag) (>-fun-ordering? (caar cl1) (caar cl2)))
;	   (cons (car cl1)
;		 (sort-merge-clauses (cdr cl1) cl2 #f) ))
;	(else
;	   (cons (car cl2)
;		 (sort-merge-clauses cl1 (cdr cl2) pair-flag) ))))




(define (make-predclause-to-propclauses fullpredclause domainsize flatflag)
  (let* ((info (car fullpredclause))
         (propclausepos '())
         (propclauseneg '()) 
         (curlit '())
         (litvalue '())
         (curproplit '())
         (poscurproplit '())
         (varslist (if flatflag
                       (get-fullclause-varslist fullpredclause)
                       (get-fullclause-varslist fullpredclause) ))
         (varvaluations '())
         (maxvar (- domainsize 1))
         (allchecked #f)
         (unsatisfiable #f)
         (tmp '())
         (tautology #f)
         (clausepart '()) )
     
    ;(newline)
    ;(display "predclause raw: ")(display fullpredclause)
    ; (display "predclause: ")(print-pure-clause (cdr fullpredclause))
    ;(newline)
    ;(display "varslist: ")(display varslist)
    ;(newline) 
     
    ; if vars exist, create a valuation list for vars     
    (if (not (null? varslist))
        (begin
          (set! varvaluations 
               (map (lambda(x) 0) varslist) )))
               
    ; loop over all the valuations for vars          
    (do ()           
        ((or unsatisfiable
             allchecked
             (> (- (get-check-time 1) *slow-start-time*)
	         *one-pass-time*))
	   (if (and (not unsatisfiable)(not allchecked))
               (set! *prop-time-ran-out* #t) )      
           unsatisfiable)  
      (set! propclausepos '())
      (set! propclauseneg '())
      (set-pred-var-values varslist varvaluations)  
      
      ;(display "varvaluations: ")(display varvaluations)
      
      (set! tautology #f)
      ; loop over literals in the clause
      (do ((clausepart (cdr fullpredclause) (cdr clausepart)))  
      	  ((or (null? clausepart) tautology unsatisfiable) 
      	     ; if no tautology and an empty clause generated, unsatisfiability is found 
      	     (if (and (not tautology) (null? propclauseneg) (null? propclausepos))
      	         (set! unsatisfiable #t) )
      	     ; if unsatisfiability was not found directly, try to add the clause    
      	     (if (not unsatisfiable)      	          
      	         (begin
      	           (if (not tautology)
      	               (set! tmp (add-new-prop-clause 
      	                            (cons propclauseneg propclausepos) )))
      	           (if (not tmp) (set! unsatisfiable #t)) )))
      	           
        ; (newline)(display "about to call make-...-pred-to-proplit with curlit: ")
        ; (display (car clausepart))(newline)     	           
      	           
      	(set! curlit (car clausepart))  
      	(if flatflag
      	    (set! curproplit (make-flat-pred-to-proplit curlit)) 
      	    (set! curproplit (make-pred-to-proplit curlit)) )
      	
      	       
       (if *prop-debug*
           (begin (newline)(display "curproplit generated: ")(display curproplit) 
                  (newline)(display "curproplit pos value so far: ")
                  (display 
                   (vector-ref *propvar-enqued-value* 
                               (if (< curproplit 0)
      	                           (- 0 curproplit) 
      	                           curproplit))))) 
      	
      	; check found curproplit: equality, subs, cut, tautology, neg, pos?
      	(cond
      	  ((eq? curproplit #t) ; a clause is a tautology
      	     (set! tautology #t) )
      	  ((eq? curproplit #f)) ; do not add a literal to a clause!      
      	  ((begin (if (< curproplit 0)
      	              (set! poscurproplit (- 0 curproplit)) 
      	              (set! poscurproplit curproplit) )      	  
      	          (set! litvalue (vector-ref *propvar-enqued-value* poscurproplit))
      	          ; (set! litvalue '())  
      	          (not (null? litvalue)) )
      	      ; subsumed clause or cut literal    
      	      (if (eq? #t litvalue)
      	          ; if same polarity, mark subsumed, otherwise do not add lit (cut)
      	          (if (>= curproplit 0)     	              
      	              (set! tautology #t) ) 
      	          (if (< curproplit 0)
      	              (set! tautology #t) )))     
      	  ((< curproplit 0)      	   
      	        (set! tmp poscurproplit)
      	        (if (memq tmp propclausepos)
      	            (set! tautology #t)
      	            (if (not (memq tmp propclauseneg)) 
      	                (set! propclauseneg (cons tmp propclauseneg)) )))
      	                     ; (add-prop-lit-to-clause tmp propclauseneg)
      	  (else
      	        (if (memq curproplit propclauseneg)
      	            (set! tautology #t)
      	            (if (not (memq curproplit propclausepos))
      	                (set! propclausepos (cons curproplit propclausepos)) ))))) 
      	                     ; (add-prop-lit-to-clause curproplit propclausepos)  
      ; take a next valuation 	  
      
      (if *prop-debug*
          (begin (newline)(display "maxvar: ")(display maxvar)(newline)
      		 (display "varvaluations: ")(display varvaluations)(newline) ))
            
      (set! allchecked 
          (or (null? varvaluations)
              (valuation-list-nextvalues varvaluations #t domainsize) )))))    	  

(define (make-extra-flatten-propclauses domainsize)
   (let ((tmp '())
         (res '())
         (wholeresult '())
         (closedaxiom '())      
         (funinfo '())         
         (var 0)
         (args '())
         (funinfolist '()) )
     ; create the input: funs and constants with their increased arities    
     (set! funinfolist 
         (append
       	    (map (lambda(x)
       		   (set! tmp (assq (car x) *prop-orig-funs-enum*))
       		   (cons (cdr tmp)(+ 1 (cdr x))))
       	         *prop-orig-funs*)
            (map (lambda(x)
       		   (set! tmp (assq x *prop-orig-constants-enum*))
       	           (cons (cdr tmp) 1))    
       		 *prop-orig-constants*)))
     ; loop over all funs and constants in the infolist 		 
     (do ((lstpart funinfolist (cdr lstpart)))
         ((null? lstpart) wholeresult)
       (set! funinfo (car lstpart))     
       (set! args '())      
       ; create arglist main part for a concrete fun, one shorter than actual
       (do ((x 1 (+ 1 x)))
           ((>= x (cdr funinfo)))
         (set! args (cons x args)) )       
       ; create closed and total axiom, where constants have been added 100000
       (set! closedaxiom '())  
       (do ((y 0 (+ 1 y)))
           ((>= y domainsize))
         (set! tmp (cons (car funinfo) 
                         (cons (+ y 100000) args) ))  
         (set! closedaxiom (cons tmp closedaxiom)) )
       (set! closedaxiom (cons '() closedaxiom))  
       (if (not (null? closedaxiom))
           (set! wholeresult (cons closedaxiom wholeresult)) )
       ; create the well-definedness axioms     
       
       ; (newline)(display "welldefinedness for funinfo:") (display (car funinfo))
       
       (do ((u 0 (+ 1 u)))
           ((>= u domainsize))
          
         ; (newline)(display u)(display  ": ")  
           
         (set! res '())  
         (do ((v 0 (+ 1 v)))
             ((>= v u))                            
                    
           ; (display v)(display " ")        
                    
           (set! res 
             (list '()
                   (cons (- 0 (car funinfo)) (cons (+ v 100000) args))
         	   (cons (- 0 (car funinfo)) (cons (+ u 100000) args)) ))
         	   
            (if (not (null? res))	   
                (set! wholeresult (cons res wholeresult)) ))))))
         	   
      
                 
(define (add-prop-lit-to-clause lit clause)
    (cond ((null? clause) (cons lit '()))          
          ((<= lit (car clause)) (cons lit clause))
          (else (cons (car clause) 
                      (add-prop-lit-to-clause lit (cdr clause)) ))))
     

(define (add-new-prop-clause propclause)
  (let ((tmp #t))
    (if *prop-debug*
         (begin (newline)(display propclause)(newline)) )     
   ; if unit clause, assign the literal value     
   (cond
     ((and (null? (car propclause))
           (null? (cddr propclause)) )      
         (vector-set! *propvar-enqued-value* (cadr propclause) #t) )
     ((and (null? (cdr propclause))
           (null? (cdar propclause)) )      
         (vector-set! *propvar-enqued-value* (caar propclause) #f) ))    
        ; add clause to the discrimination tree         
        ;(set! tmp (add-discr-tree-propclause (car propclause) (cdr propclause)))                     
        ;(set! tmp (check-exist-propclause (car propclause) (cdr propclause))) 
   ; add clause if not subsumed       
   (set! *propclauses* (cons propclause *propclauses*))
   #t))         


(define (check-exist-propclause neglits poslits)
  (let ((found #f))
   (do ((part *propclauses* (cdr part)))
       ((or found (null? part)) 
         ;(display found)
         (not found)) 
    
    ;(newline)
    ;(display "car part: ")(display (car part))
    ;(display " neg: ")(display neglits)(newline) 
    ;(display " pos: ")(display poslits)(newline)   
       
    (if (and (sub-ordered-list? (caar part) neglits)
             (sub-ordered-list? (cdar part) poslits)  )
        (set! found #t) ))))       


(define (sub-ordered-list? slist glist)
  (or (null? slist)
      (and (not (null? glist)) 
           (or (and (eq? (car slist) (car glist))
           	    (sub-ordered-list? (cdr slist) (cdr glist)) )
	       (and (> (car slist) (car glist))           	    
	            (sub-ordered-list? slist (cdr glist)) )))))



; ((1 (-2 #t))
;
;
;


;(define (add-discr-tree-propclause tree neglits poslits)
;   (let* ((treepart '())
;          (tmp '()) 
;          (nneq neglits)
;          (npos poslits)
;          (nlit '())
;     (if (null? neglits) 
;         (if (null? poslits)
;             '()
;             (begin (set! npos (cdr poslits))             
;                    (set! nlit (car poslits)) ))
;         (begin (set! nneg (cdr neglits))
;                (set! nlit (- 0 (car neglits))) ))            
;     (set! treepart (assq nlit (cdr tree)))
;     (if treepart
;        (set-cdr! tree
;                (cons 
;                   (add-discr-tree-propclause tree nneg npos)
;                   (cdr tree)))))))
     
         
(define (get-fullclause-varslist fullpredclause)
   (set! *tmp-varlist* '())
   (for-each (lambda(lit)(get-fullclause-varslist-aux! lit)) 
             (cdr fullpredclause) )
   *tmp-varlist*)          


(define (get-fullclause-varslist-aux! term) 

    (if *prop-debug*
          (begin (newline)(display "term: ")(display term)
    		 (newline) (display "fun-may-cont-vars: ")
    		 (display (fun-may-cont-vars? (car term)))(newline) ))
    
    (do ((part (cdr term) (cdr part)))
	((null? part))
      (set! term (car part))
      
    (if *prop-debug*
          (begin  
    		(newline)(display "inner term: ")(display term)    
    		(newline) (display "fun-may-cont-vars: ")
    		(if (pair? term)(display (fun-may-cont-vars? (car term))))(newline) ))
    
      
      (if (non-composite? term)
	  (if (non-composite:-var? term)
	       (if (not (memq (var-nr term) *tmp-varlist*))
	           (set! *tmp-varlist* (cons (var-nr term) *tmp-varlist*) )))	     
	  (if #t ; (fun-may-cont-vars? (car term))
	      (get-fullclause-varslist-aux! term) ))))


(define (set-pred-var-values varslit varvaluations)  

   (if *prop-debug*
          (begin
   		(newline)(display "varslit: ")(display varslit)
   		(newline)(display "varvaluations: ")(display varvaluations)
   		(newline) ))

   (for-each (lambda(var val)(set-var-ref! var val))
              varslit
              varvaluations))
  
(define (make-pred-to-proplit curlit)
  (let* ((negflag (neg-fun? (car curlit)))
         (purefun (pure-unsigned-fun (car curlit)))
         (arg1 '())
         (arg2 '())
         (treepart '())
         (res '()) )
         
       (if *prop-debug*
          (begin
       	   (display "curlit: ")(display curlit)(newline)  
       	   (display "*prop-lit-mappings*: ")(display *prop-lit-mappings*)(newline) ))
     
      (if (pos-eq-pred? purefun)      	    
        ; equality case
        (begin
          (if (eq? (get-pred-term-interpvalue (cadr curlit))
                   (get-pred-term-interpvalue (caddr curlit)) )
              (if negflag (set! res #f) (set! res #t))
              (if negflag (set! res #t) (set! res #f)) )
          #f)
        ; standard lit case  
        (begin                 
         (set! treepart (assq purefun *prop-lit-mappings*))      
         (if *prop-debug*
             (begin (display "treepart: ")(display treepart)(newline)) )
      
         (if (not treepart)
             (begin (display "error in pred-to-prop pred evaluation")(quit)) )  
         (set! treepart (make-pred-to-proplit-aux (cdr curlit) (cdr treepart)))
         (if negflag
             (set! res (- 0 treepart))
             (set! res treepart) )))
     res))    



(define (make-flat-pred-to-proplit curlit)
  (let* ((negflag (< (car curlit) 0))
         (purefun (if negflag (- 0 (car curlit)) (car curlit)))
         (arg1 '())
         (arg2 '())
         (treepart '())
         (res '()) )
         
       (if *prop-debug*
          (begin
       	   (display "curlit: ")(display curlit)(newline)  
       	   (display "*prop-lit-mappings*: ")(display *prop-lit-mappings*)(newline) ))
     
      (if (eq? *prop-flat-eq-symb* purefun)    
        ; equality case
        (begin
          (if (eq? (get-pred-term-interpvalue (cadr curlit))
                   (get-pred-term-interpvalue (caddr curlit)) )
              (if negflag (set! res #f) (set! res #t))
              (if negflag (set! res #t) (set! res #f)) )
          #f)
        ; standard lit case  
        (begin                 
         (set! treepart (assq purefun *prop-lit-mappings*))      
         (if *prop-debug*
             (begin (display "treepart: ")(display treepart)(newline)) )
      
         (if (not treepart)
             (begin (display "error in pred-to-prop pred evaluation")(quit)) )  
         (set! treepart (make-flat-pred-to-proplit-aux (cdr curlit) (cdr treepart)))
         (if negflag
             (set! res (- 0 treepart))
             (set! res treepart) )))
     res))    

     
; check arglist of a fun     
     
(define (make-pred-to-proplit-aux funargs treepart)
  (let ((tmp '())      
        (argval '()) )
    ; cycle over all funargs    
    (do ((argspart funargs (cdr argspart)))
       ((null? argspart) treepart)   
     ; evaluate interpreted funarg
     (set! argval (get-pred-term-interpvalue (car argspart)))
         
     ; go further in the tree
     
     (if *prop-debug*
          (begin
     		(display "funargs: ")(display funargs)(newline)  
     		(display "argval: ")(display argval)(newline)
     		(display "treepart: ")(display treepart)(newline) ))
      
     (set! tmp (assq argval treepart))
     
     
     (if (not tmp)
         (begin (display "error in pred-to-prop fun evaluation")(abort)) )
     (set! treepart (cdr tmp)) )))

(define (make-flat-pred-to-proplit-aux funargs treepart)
  (let ((tmp '())
        (arg '())
        (argval '()) )
    ; cycle over all funargs    
    (do ((argspart funargs (cdr argspart)))
       ((null? argspart) treepart)
     (set! arg (car argspart))   
     ; evaluate interpreted funarg   
     (if (<= 100000 arg) 
         (set! argval (- arg 100000)) ; constant in special new axioms
         (set! argval (var-ref (var-nr arg))) ) ; variable
     
     ; go further in the tree
     
     (if *prop-debug*
          (begin
     		(display "argspart: ")(display argspart)(newline)  
     		(display "argval: ")(display argval)(newline)
     		(display "treepart: ")(display treepart)(newline) ))
      
     (set! tmp (assq argval treepart))
     
     
     (if (not tmp)
         (begin (display "error in pred-to-prop fun evaluation")(abort)) )
     (set! treepart (cdr tmp)) )))
     
; check a full term     
     
(define (get-pred-term-interpvalue term)
  (let ((tmp '())
        (treepart '()) ) 
    (if (non-composite? term)
        (if (non-composite:-var? term) 
            (var-ref (var-nr term))            
            (begin 
              (set! tmp (assq term *prop-constants-mapping*))
              (if (not tmp)
                  (begin (display "error in pred-to-prop fun evaluation")(quit)) )
              (cdr tmp) ))
        (begin           
          (set! treepart (assq (pure-fun (car term)) *prop-fun-mappings*))
          (if (not treepart)
              (begin (display "error in pred-to-prop fun evaluation")(quit)) )
          (make-pred-to-proplit-aux (cdr term) (cdr treepart)) ))))


   
;----------------------------------------------------------------------
; 
;       flattening a first-order formula
;
;----------------------------------------------------------------------
;
; each first-order clause is first converted to a simple form with
; numbers representing predicate and function symbols, plus constants
; represented as 0-argument function applications, plus vars represented
; as integers.
;
; afterward n clauses are built from the converted clause
;
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define *prop-flatten-lastvar* 0)
(define *prop-flatten-newlits* '())
(define *prop-flatten-varslist* '())
(define *prop-flatten-clause-processed-terms* '())

(define (flatten-clause-set clauses)
 (let ((tmp '())
       (clause '())
       (res '()))
   (do ((clpart clauses (cdr clpart)))
       ((null? clpart)   
          res)
     (set! clause (car clpart))  
     (if (and (pair? clause)(pair? (cdr clause)))
         (begin
           
            ;(newline)(display "about to flatten clause: ")
            ;(print-clause clause)(newline)
           
         
           (set! tmp (flatten-clause clause))
           
            ;(display "flattening gives: ")(newline)           
            ;(for-each (lambda(x)(display x)(newline)) tmp)
           
           (set! res (append tmp res)) )))))  


; flatten-clause takes a predclause and returns the set of fully flattened
; clauses, having at the head the list of all vars in the new clause

(define (flatten-clause clause)
   
   ; (display "starting to flatten clause: ")(display clause)(newline)

   (let ((preprocessed '()))
     (set! *prop-flatten-lastvar* 100)
     (set! preprocessed (preprocess-flatten-clause (cdr clause)))
     (if (>= *prop-flatten-lastvar* 150)
         (begin (display "too large clause for flattening satisfiability check")(quit)) )
     
     ;(newline)(display "flatten-clause gave preprocessed set: ")(newline)
     ;(for-each (lambda(x)(display x)(newline)) preprocessed)
                
     
     (map (lambda(cl)(flatten-clause-aux cl)) 
          preprocessed) ))

; preprocess-flatten-clause creates a list of clauses from a single clause, splitting
; all the positive equalities

(define (preprocess-flatten-clause pureclause)
  (let ((lit '())
        (fun '())
        (purefun '())
        (negfun '())
        (negflag #f)
        (eqflag #f)  
        (var '())
        (l1 '())
        (l2 '())
        (tmp '()))
        
    ;(newline)(display "preprocess-flatten-clause called with: ")(newline)        
    ;(display pureclause)(newline)
        
    (if (null? pureclause)
        (list '())
        (begin
          (set! lit (car pureclause))
          (set! fun (car lit))
          (set! purefun (pure-unsigned-fun fun))
          (set! negflag (neg-fun? fun))
          (set! eqflag (pos-eq-pred? purefun))  
          (if (and eqflag
                   (not negflag)
                   (not (and (non-composite? (cadr lit)) ; not var?
                             (non-composite:-var? (cadr lit)) )) 
                   (not (and (non-composite? (caddr lit)) ; not var?
                             (non-composite:-var? (caddr lit)) ))) 
              ; both args are non-var               
              (begin 
              
                 ; (newline)(display "both args non-var")(newline)
              
                 (set! var (make-prop-flatten-newvar))
                 (set! l1 (list purefun (cadr lit) var))
                 (set! l2 (list purefun (caddr lit) var))
                 (set! negfun (logior *sgn-mask* purefun))
                 (set! tmp (preprocess-flatten-clause (cdr pureclause)))
                 
                 ; (newline)
                 ; (display "var: ")(display var)(newline)
                 ; (display "l1: ")(display l1)(newline)
                 ; (display "l2: ")(display l2)(newline)
                 ; (display "negfun: ")(display negfun)(newline)
                 ; (newline)
                 
                 
                 (append (map (lambda(cl)(cons l1 (cons (cons negfun (cdr l2)) cl)))
                              tmp)
                         (map (lambda(cl)(cons (cons negfun (cdr l1)) (cons l2 cl)))
                              tmp)))     
              
              ; either not equality or at least one of the vars was var
              (begin
              
                 ; (newline)(display "not both args non-var")(newline)
                                
                 (set! tmp (preprocess-flatten-clause (cdr pureclause)))
                 (map (lambda(cl)(cons lit cl)) tmp) ))))))
            


; flatten-clause-aux takes a full predclause WITHOUT SPLITTABLE EQUALITY 
; and returns the fully flattened
; clause, having at the head the list of all vars in the new clause.


(define (flatten-clause-aux clause)
 (let ((lit '())
       (fun '())
       (purefun '())
       (negflag #f)
       (eqflag #f)
       (varslist (get-fullclause-varslist clause))
       (res '())
       (processed-lit '()) )      
       
   ; (newline)(display "original varslist: ")(display varslist)(newline)    
       
   (if (> (length varslist) 40)
       (begin
         (display "too many vars (over 40) in the clause for the sat checker")
         (quit) ))
   (set! *prop-flatten-varslist* varslist)      
   (set! *prop-flatten-lastvar* 150)
   (set! *prop-flatten-newlits* '())      
   (set! *prop-flatten-clause-processed-terms* '())
   (do ((clpart clause (cdr clpart)))
       ((null? clpart)
         (if (> *prop-flatten-lastvar* 250)
             (begin
              (display "too many vars (over 250) in the clause for the sat checker")
              (quit) ))
           (cons '() (append (reverse *prop-flatten-newlits*) (reverse res))) )    
       ;  (cons
       ;    *prop-flatten-varslist*           
       ;     (append (reverse *prop-flatten-newlits*) (reverse res)) ))          
     (set! processed-lit (flatten-composite-term (car clpart) #t))  
     (set! res (cons processed-lit res)) )))
          

(define (flatten-composite-term term litflag)   
 (let ((funsymb '())
       (fun '())  
       (negflag #f)
       (eqflag #f)
       (purefun '())
       (tmp '()) 
       (newargs '())
       (newvar '()) )      
       
   ; (newline)(display "about to flatten comp term: ")(display term)     
    
  ; first process the leading symbol of the term  
  (if (not litflag)
    ; process leading function symbol
    (begin     
        (set! funsymb
             (cdr (assq (pure-fun (car term)) *prop-orig-funs-enum*)) )        
        (set! term (cons funsymb (cdr term))) )     
    ; process predicate symbol
    (begin
        (set! fun (car term))
        (set! purefun (pure-unsigned-fun fun))
        (set! negflag (neg-fun? fun))
        (set! eqflag (pos-eq-pred? purefun))    
        (set! funsymb (cdr (assq purefun *prop-orig-preds-enum*)))   
        
        ; (newline)
        ; (display "purefun: ")(display purefun)(newline)
        ; (display "eqflag: ")(display eqflag)(newline)
        ; (display "negflag: ")(display negflag)(newline)
        ; (newline)
        
        
        (cond 
          ; not an equality literal
          ((not eqflag)                      
             (if negflag (set! funsymb (- 0 funsymb)))           
             (set! term (cons funsymb (cdr term))) )             
          ; eq, both args vars, handle normally as eq
          ((and eqflag
                (non-composite? (cadr term))
                (non-composite:-var? (cadr term))
                (non-composite? (caddr term))
                (non-composite:-var? (caddr term)) )
             (if negflag (set! funsymb (- 0 funsymb)))           
             (set! term (cons funsymb (cdr term))) )                    
             ; (display "equality has two var args in flatten-composite-term")(abort) 
          ; eq, first arg var?   
          ((and eqflag
                (non-composite? (cadr term))
                (non-composite:-var? (cadr term)) )                    
            (if (non-composite? (caddr term))
                (begin
                  (set! tmp (assq (caddr term) *prop-orig-constants-enum*))
                  (set! term
                        (cons (if negflag (- 0 (cdr tmp)) (cdr tmp))
                              (cons (cadr term) '())) ))
                (begin
                  (set! tmp (assq (pure-fun (car (caddr term))) *prop-orig-funs-enum*))
                  (set! term 
                        (cons (if negflag (- 0 (cdr tmp)) (cdr tmp)) 
                              (cons (cadr term) (cdr (caddr term)))) ))))             
          ; eq, second arg var?   
          ((and eqflag
                (non-composite? (caddr term))
                (non-composite:-var? (caddr term)) ) 
                
            ; (display "second eq arg var")(newline)    
                
            (if (non-composite? (cadr term))
                (begin
                  (set! tmp (assq (cadr term) *prop-orig-constants-enum*))
                  (set! term
                        (cons (if negflag (- 0 (cdr tmp)) (cdr tmp)) 
                              (cons (caddr term) '())) ))
                (begin
                  (set! tmp (assq (pure-fun (car (cadr term))) *prop-orig-funs-enum*))
                  (set! term 
                        (cons (if negflag (- 0 (cdr tmp)) (cdr tmp))
                              (cons (caddr term) (cdr (cadr term)))) ))))                        
          ; negative eq, none of the args var          
          ((and eqflag negflag)            
             (set! term (cons (- 0 funsymb) (cdr term))) )          
          ; positive eq, none of the args var: ERROR!
          (else                  
             (display "equality has two non-var args in flatten-composite-term")
             (abort) ))))
    

   ; (newline)(display "term after funsymb-processing: ")(display term)        

       
  (set! newargs
   (map 
    (lambda(curterm)    
     (if (non-composite? curterm)
         (if (non-composite:-var? curterm) 
            ; (var-ref (var-nr curterm))            
            (set! tmp (var-nr curterm))
            (begin 
              ; test if an equal curterm already processed in the clause
              (set! tmp (assoc curterm *prop-flatten-clause-processed-terms*))
              (if tmp
                  ; an equal term already processed, use assigned var
                  (set! tmp (cdr tmp))
                  ; not processed yet, create a new var
                  (begin
              	    (set! tmp (assq curterm *prop-orig-constants-enum*))
                    (if (not tmp)
                        (begin (display "error in pred-to-prop fun evaluation")(quit)) )
                    (set! newvar (make-prop-flatten-newvar))    
                    (make-prop-flatten-addpair curterm (list (cdr tmp)) newvar)
                    (set! tmp newvar) ))))
        (begin
          ; test if an equal term already processed in the clause
          (set! tmp (assoc curterm *prop-flatten-clause-processed-terms*))
          (if tmp
              ; an equal term already processed, use assigned var
              (set! tmp (cdr tmp))
              ; not processed yet, create a new var
              (begin                                 
                (set! tmp (flatten-composite-term curterm #f))
                (set! newvar (make-prop-flatten-newvar))    
                (make-prop-flatten-addpair curterm tmp newvar)
                (set! tmp newvar) ))))
      tmp)
    (cdr term) ))
    
   ; (newline)(display "newargs found for the term: ")(display newargs)                               
             
   ; compose the whole        
   (cons (car term) newargs) )) 
    

(define (make-prop-flatten-newvar)
  (set! *prop-flatten-lastvar* (+ 1 *prop-flatten-lastvar*))
  (set! *prop-flatten-varslist* (cons *prop-flatten-lastvar* *prop-flatten-varslist*))
  *prop-flatten-lastvar*)

  
(define (make-prop-flatten-addpair origterm newterm newvar)
  (let ((newlit '()))
     (set! *prop-flatten-clause-processed-terms*
           (cons (cons origterm newvar) *prop-flatten-clause-processed-terms*) )           
     (set! newlit (cons (- 0 (car newterm)) 
                        (cons newvar 
                              (cdr newterm) )))
     (set! *prop-flatten-newlits*
          (cons newlit *prop-flatten-newlits*) )))
    
   
;----------------------------------------------------------------------
; 
;       prop formula solving: unsatisfiable or not?      
;
;----------------------------------------------------------------------
;
;  variable:
;   an integer, pointing to four vectors:
;   *propvar-value*                   vector containing (), #t or #f
;   *propvar-negative-occ-clauses*    vector pointing to lists of clauses
;   *propvar-positive-occ-clauses*    vector pointing to lists of clauses
;   *propvar-enqued-value*            vector 
;  
;  clause:
;
;   propclause-nr-of-negative-active
;   propclause-nr-of-positive-active
;   negative-literals   
;   positive-literals
;   subsumer
;
;  
;  clause list:
;
;  queue:
;   *prop-var-assignment-queue* 
; 
;  backtrace data: 
;   *prop-assigned-varvalpairs*
;   *prop-unsat-found-occlistpart*
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define *prop-debug* #f)

(define *propvar-value*  '())      
(define *propvar-negative-occ-clauses* '())    
(define *propvar-positive-occ-clauses* '())    
(define *propvar-enqued-value* '()) 
(define *prop-var-assignment-queue* '())
(define *prop-assigned-varvalpairs* '()) 
(define *prop-unsat-found-occlistpart* '()) 

; create-prop-var-vectors creates necessary propvar vectors

(define (create-prop-var-vectors maxvar)
  (set! *propvar-value* (make-vector (+ 1 maxvar) '()))
  (set! *propvar-positive-occ-clauses* (make-vector (+ 1 maxvar) '()))
  (set! *propvar-negative-occ-clauses* (make-vector (+ 1 maxvar) '()))
  (set! *propvar-enqued-value* (make-vector (+ 1 maxvar) '())) )

; initialize-prop-var-values sets the values of vars to initial null status

(define (initialize-prop-var-values maxvar)
  (vector-fill! *propvar-value* '())
  (vector-fill! *propvar-positive-occ-clauses* '())
  (vector-fill! *propvar-negative-occ-clauses* '())
  (vector-fill! *propvar-enqued-value* '()) )
  
 ;(do ((var 0 (+ 1 var)))
 ;     ((> var maxvar)) 
 ;  (vector-set! *propvar-value* var '())
 ;  (vector-set! *propvar-positive-occ-clauses* var '())
 ;  (vector-set! *propvar-negative-occ-clauses* var '())
 ;  (vector-set! *propvar-enqued-value* var '()) ))

; build-prop-internal-structure creates an ldpp-usable clause list
; and sets the variable vectors to point to right values
;
; input clauses are assumed to be a list of clauses, where each clause
; is a pair of two lists: positive occ var nrs and negative occ var nrs.


(define (build-prop-internal-structure clauses)
  (let ((resclauses '())
        (tmp '())
        (var '()) 
        (repeatflag #f)
        (resclause '())
        (neglits '())
        (poslits '()) )        
    (do ((part clauses (cdr part)))
        ((null? part) resclauses)       
      (set! poslits (cdar part))
      (set! neglits (caar part))
      
      (if *prop-debug*
          (begin (display poslits)(newline)(display neglits)(newline)) )
      
      ; create a new clause
      (set! resclause 
           (list (length neglits) (length poslits) neglits poslits '()) )
      (set! resclauses (cons resclause resclauses))              
      ; loop over poslits, set variable values    
      (do ((part poslits (cdr part)))
          ((null? part))
        (set! var (car part))       
        (vector-set! *propvar-positive-occ-clauses* var
           (cons resclause (vector-ref *propvar-positive-occ-clauses* var)) ))
      ; loop over neglits, set variable values    
      (do ((part neglits (cdr part)))
          ((null? part))
        (set! var (car part))       
        (vector-set! *propvar-negative-occ-clauses* var
           (cons resclause (vector-ref *propvar-negative-occ-clauses* var)) )))))
              

; solve-prop-clauses gives #t if satisfiable, #f if not,
;
; input clauses are assumed to be a list of clauses, where each clause
; is a pair of two lists: positive occ var nrs and negative occ var nrs.
;
; maxvar should be at least as big as the biggest var nr. OK to be bigger.
;
; variable value vectors must be created before.

(define (solve-prop-clauses clauses maxvar)
  (let ((splittableclauses '())
        (newsplitvar '())
        (internalclauses '())
        (res '()) ) 
    (initialize-prop-var-values maxvar)
    
    (if *prop-debug*  (begin 
      (newline)(display "-- prop values initialised")(newline)
      (show-prop-internal-data '() maxvar) ))
    
    (set! internalclauses (build-prop-internal-structure clauses))   
    (set! *prop-internal-clauses* internalclauses)
    (set! *prop-maxvar* maxvar)
    
    (if *prop-debug*  (begin 
      (newline)(display "-- prop internal structure built")(newline)       
      (show-prop-internal-data internalclauses maxvar) ))
    
    (if (or *mace-model-strat*
            (null? *prop-orig-funs*)
            (and *prop-unsat-strat* *simple-model-strat*) )
        (begin
          (newline)
          (display (length  clauses))
          (display " propositional clauses containing ")
          (display maxvar)
          (display " variables were built.") 
          (newline) ))  
            
            
    
    (if (> *display-info-level* 90) 
           (begin (newline)
                  (display "clauses for full prop evaluation: ")
                  (newline)
	          (for-each (lambda(cl)(display cl)(newline)) *prop-internal-clauses*)    
		  (newline) ))
    
    (if #f 
     (begin
       (newline)(newline)(newline) 
       (for-each 
         (lambda(cl) 
           (for-each (lambda(n)
                      (display "-")(display "p")(display n)(display " | ") )
                     (caddr cl) )
           (for-each (lambda(p)
                      (display "p")(display p)(display " | ") )
                     (cadddr cl) )
          (display ".")
          (newline) )          
         *prop-internal-clauses*)                        
       (newline)(newline)(newline) ))
    
     
    (set! *prop-var-assignment-queue* '())    
    (set! *prop-assigned-varvalpairs* '())
    (set! *prop-unsat-found-occlistpart* '())  
    ; try simple propagation of all units first
    (set! res (initial-prop-unit-propagate internalclauses)) ; if returns #t, found unsat
    
    (if *prop-debug* 
     (begin 
      (newline)(display "-- initial-prop-unit-propagate finished:")(newline)       
      (display "  satisfiable?: ")(display (not res))(newline)
      (show-prop-internal-data internalclauses maxvar) ))
    
    (if res 
        #f ; unsatisfiable
        (begin
          ; compute splittable 
          (set! splittableclauses (prop-find-splittable internalclauses))
          (set! newsplitvar (prop-give-next-split splittableclauses))
          
          (if *prop-debug*  (begin 
            (newline)(display "-- newsplitvar: ")(display newsplitvar)(newline) ))
          
          (if (null? newsplitvar)
              #t ; satisfiable, since could not split
              (or (solve-prop-clauses-branch 
                       newsplitvar #t splittableclauses)
                  (solve-prop-clauses-branch 
                       newsplitvar #f splittableclauses)))))))
   
; initial-prop-unit-propagate finds and propagates all the 
; initially unit clauses in the input set of clauses
; if contradiction found, it returns #t

(define (initial-prop-unit-propagate clauses)
  (let ((tmp '())
        (tmp2 '())
        (res #f) )
    (do ((part clauses (cdr part)))
        ((or (null? part) res) 
            (if res 
                res ; contradiction found while pushing to the queue
                ; propagate all units
                (loop-prop-unit-queue) ))                 
      (set! tmp (car part)) 
      
      (if *prop-debug*  (begin (display tmp)(newline)))
      
      (if (eq? 1 (+ (car tmp) (cadr tmp))) ; sum of active lits is 1
          (begin            
            (set! tmp2 (cddr tmp))
            (if (null? (car tmp2))
                ; negative empty, hence a positive unit
                (set! res (prop-enque-unit (caadr tmp2) #t))        
                ; negative not empty, hence a negative unit
        	(set! res (prop-enque-unit (caar tmp2) #f)) ))))))
        
   
   
; solve-prop-clauses-branch returns
; #t if satisfiable,
; #f if unsatisfiable
; assignments etc are not cleared, if found to be satisfiable

(define (solve-prop-clauses-branch splitvar splitvalue splittableclauses)

 (if *prop-debug*  (begin 
   (newline)(display "-- solve-prop-clauses-branch called: ")(newline)       
   (display "  splitvar, splitvalue: ")
   (display splitvar)(display " ")(display splitvalue)
   (newline) ))

  (let ((res '())
        (newsplitvar '()) 
        (local-assignment-queue '())
        (local-assigned-varvalpairs '()) )         
    (set! *prop-time-ran-out* #f)   
    (set! *prop-var-assignment-queue* '())    
    (set! *prop-assigned-varvalpairs* '())    
    (set! *prop-unsat-found-occlistpart* '())  
    (set! *prop-var-assignment-queue* 
          (cons (cons splitvar splitvalue) '()) )  
                       
    ;;(set! test-list
    ;;  (map (lambda(cl)(cons cl (cons (car cl) (cadr cl))))
    ;;       *prop-internal-clauses*))                     
                    
    (set! res (loop-prop-unit-queue)) ; if returns #t, found unsat
    (if res
        ; found to be unsat by simple propagation 
        (begin
           ; undo assignment and propagation results                                
           (prop-undo-assignments *prop-assigned-varvalpairs*)
           (prop-undo-queue-assignments *prop-var-assignment-queue*)
           
           ;;(test-prop-undo-assignments *prop-assigned-varvalpairs* )
           ;;(prop-undo-queue-assignments *prop-var-assignment-queue*)                      
           ;;(for-each (lambda(back)
           ;;            (set-car! (car back) (cadr back))
           ;;            (set-car! (cdar back) (cddr back)) )
           ;;          test-list)  
           
           #f) ; return unsatisfiable
        ; not found to be unsat by simple propagation
        (if (> (- (get-check-time 1) *slow-start-time*)
	        *one-pass-time*)
          ; time ran out 
          (begin
            (set! *prop-time-ran-out* #t)
            #f)
          ; time left  
          (begin
           ; store backtracing lists locally
           (set! local-assigned-varvalpairs *prop-assigned-varvalpairs*)
           (set! *prop-assigned-varvalpairs* '())                       
           (set! local-assignment-queue *prop-var-assignment-queue*)
           (set! *prop-var-assignment-queue* '())      
           ; propagation did not find contradiction, now try to split
           (set! newsplitvar (prop-give-next-split splittableclauses))
           
           (if *prop-debug*  (begin 
            (newline)(display "-- newsplitvar: ")(display newsplitvar)(newline) ))
           
           (if (null? newsplitvar)
               #t ; splitting not possible, now satisfiable. 
               ; try to use splitting: if one branch gives #t, satisfiable
               (if (or (solve-prop-clauses-branch 
                         newsplitvar #t splittableclauses)
                       (solve-prop-clauses-branch 
                         newsplitvar #f splittableclauses))
                    ; one branch satisfiable, hence branch satisfiable
                    #t 
                    ; branch unsatisfiable, undo assignments
                    ; in simple propagation at the beginning of branch 
                    (begin                     
                      (set! *prop-unsat-found-occlistpart* #f)                 
                      (set! *prop-assigned-varvalpairs* local-assigned-varvalpairs)                     
                      (prop-undo-assignments local-assigned-varvalpairs)
                      (prop-undo-queue-assignments local-assignment-queue)        
                                            
                      ;;(test-prop-undo-assignments local-assigned-varvalpairs) 
                      ;;(prop-undo-queue-assignments local-assignment-queue)                                	                          
           	      ;;(for-each (lambda(back)
                      ;; 			(set-car! (car back) (cadr back))
                      ;; 			(set-car! (cdar back) (cddr back)) )
                      ;;         test-list)  
                     
                      #f))))))))     
                                            

; prop-find-splittable selects all clauses with more than two positive
; literals active

(define (prop-find-splittable clauses)
  (let ((res '()))
    (do ((part clauses (cdr part)))
        ((null? part) res)
      (if (> (cadar part) 1) (set! res (cons (car part) res))) )))  
    

; prop-give-next-split returns a var, if possible,
; if no splittable vars, it returns ()

(define (prop-give-next-split splittableclauses) 

 (if *prop-debug* (begin 
   (newline)(display "-- prop-give-next-split-called, splittable: ")(newline)       
   (for-each (lambda(cl)(display "  ")(display cl)(newline)) splittableclauses)
   (newline) ))

 (let ((bestclause '())
       (bestlen 100000) 
       (foundflag #f)
       (foundlit #f)
       (tmp '()) )
  (do ((part splittableclauses (cdr part)))
      ((or (null? part) foundflag) 
           (if (null? bestclause)
               bestclause
               ; give the fist unassigned positive lit of the found clause       
               (do ((litpart (car (cdddr bestclause)) (cdr litpart)))
                   ((or (null? litpart) foundlit) foundlit)
                 (if (null? (vector-ref *propvar-value* (car litpart)))
                     (set! foundlit (car litpart)) ))))
     ; consider only fully positive nonunit, not subsumed clauses,
     ; find shortest
     (if (and (eq? 0 (caar part)) ; fully positive
              (begin (set! tmp (cadar part)) (> tmp 1)) ; nonunit
              (null? (car (cddddr (car part)))) ; not subsumed
              (or (eq? 2 tmp) (< tmp bestlen)) )
         ; might be a good candidate, check subsumption further
         (if (not (prop-clause-subsumed-by-var? (car part)))
             (begin
                (set! bestclause (car part))
                (set! bestlen tmp) 
                (if (eq? 2 tmp) (set! foundflag #t)) ))))))
                     

 
(define (prop-clause-subsumed-by-var? clause)
  (let ((subsflag #f))
    ; check negative literals
    (do ((part (car (cddr clause)) (cdr part)))
        ((or (null? part) subsflag) subsflag)
      (if (eq? (vector-ref *propvar-value* (car part)) #f)
          (set! subsflag #t) ))   
    ; check positive literals  
    (if subsflag
        subsflag
        (do ((part (car (cdddr clause)) (cdr part)))
            ((or (null? part) subsflag) subsflag)
          (if (eq? (vector-ref *propvar-value* (car part)) #t)
              (set! subsflag #t) )))))  
            

; loop-prop-unit-queue loops over the queue, assigns values,
; propagates, etc until either the queue is empty or contradiction
; is found
; iff contradiction is found, it returns #t, else #f


(define (loop-prop-unit-queue)
  
  (if *prop-debug*  (begin 
    (newline)(display "-- loop-prop-unit-queue called")(newline)       
    (show-prop-internal-data '() 4)  ))

  (let ((unsat-flag #f)
        (tmp '()) )   
    (do ()
        ((or (null? *prop-var-assignment-queue*) unsat-flag) unsat-flag)
      (set! tmp (car *prop-var-assignment-queue*))
      (set! *prop-var-assignment-queue* (cdr *prop-var-assignment-queue*)) 
      (set! unsat-flag (assign-prop-var tmp)) )))


; assign-prop var returns #t iff unsatisfiable else #f
; it calls prop-resolve-clause when propagating the value

(define (assign-prop-var varvalpair)

  (if *prop-debug*  (begin 
    (newline)(display "++ assign-prop-var called: ")
    (display varvalpair)(newline) ))

  (let* ((varnr (car varvalpair))
         (varval (cdr varvalpair)) 
         (tmpres #f) 
         (unsat-flag #f) )
    ; set var value     
    (vector-set! *propvar-value* varnr varval)
    ; keep the setting for backtracing
    (set! *prop-assigned-varvalpairs* 
          (cons varvalpair *prop-assigned-varvalpairs*) ) 
    ; loop over correctly signed occurrences of var in all clauses
    (if varval
        ; if var set to #t, loop over all negative occurrences
        (do ((occlist (vector-ref *propvar-negative-occ-clauses* varnr)
                      (cdr occlist) ))
            ;((or (null? occlist) unsat-flag) unsat-flag)             
            ((null? occlist) unsat-flag)
            
          (if *prop-debug*  (begin (display (car occlist))(newline)))  
            
          (set! tmpres (prop-resolve-clause occlist varval)) 
          (if tmpres (set! unsat-flag #t)) )
        ; if var set to #f, loop over all positive occurrences   
        (do ((occlist (vector-ref *propvar-positive-occ-clauses* varnr)
                      (cdr occlist) ))   
            ;((or (null? occlist) unsat-flag) unsat-flag)      
            ((null? occlist) unsat-flag)
            
          (if *prop-debug*  (begin (display (car occlist))(newline)))  
            
          (set! tmpres (prop-resolve-clause occlist varval))
          (if tmpres (set! unsat-flag #t)) ))))
          
; prop-resolve-clause decreases the clause active propvar count and if necessary,
; puts a new element to the assignment queue
; or finds a contradiction:
; iff contradiction found, #t is returned

(define (prop-resolve-clause propclauselst varval)

  (if *prop-debug*  (begin 
    (display " ++ prop-resolve-clause called: ")(newline)
    (display propclauselst)(newline)
    (display varval)(newline) ))

  (let* ((propclause (car propclauselst))
         (clcdr (cdr propclause))
         (clsum (+ (car propclause) (car clcdr))) 
         (unsat-flag #f) )
    (if varval 
        (set-car! propclause (- (car propclause) 1))
        (set-car! clcdr (- (car clcdr) 1)) )
    (if (> clsum 2) 
        ; nothing interesting
        #f
        ; smth special happens
        (if (<= clsum 1)
            ; contradiction found
            #t
            ;;(begin
            ;;  (set! *prop-unsat-found-occlistpart* propclauselst)                                          
            ;;  #t) 
            ; unit generated: locate which unit, enque
            (begin 
               (set! varval #f) ; used as a flag for termination
               ; check negative literals
               (do ((propclause (cadr clcdr) (cdr propclause)))
                   ((or (null? propclause) varval)) 
                   
                 (if *prop-debug*  (begin   
                   (display  (car propclause))
                   (display " ")
                   (display (vector-ref *propvar-value* (car propclause)))
                   (newline) ))
                   
                 (if (null? (vector-ref *propvar-value* (car propclause)))
                     (begin 
                       (set! varval #t) 
                       (set! unsat-flag
                            (prop-enque-unit (car propclause) #f) ))))
                       ;;(if unsat-flag 
                       ;;    (set! *prop-unsat-found-occlistpart* 
                       ;;          propclauselst))
               ; check positive literals        
               (if unsat-flag
                   unsat-flag
                  (if varval
                      #f                                   
                      (do ((propclause (caddr clcdr) (cdr propclause)))
                          ((or (null? propclause) varval) unsat-flag) 
                          
                        (if *prop-debug*  (begin   
                          (display  (car propclause))
                          (display " ")
                          (display (vector-ref *propvar-value* (car propclause)))
                          (newline) ))
                          
                        (if (null? (vector-ref *propvar-value* (car propclause)))
                            (begin    
                              (set! varval #t) 
                              (set! unsat-flag
                                   (prop-enque-unit 
                                     (car propclause) #t)) ))))))))))
                                     
                              ;;(if unsat-flag 
                              ;;   (set! *prop-unsat-found-occlistpart*
                              ;;         propclauselst))

                                     
(define (prop-enque-unit varnr varval)
  (let* ((queue-val (vector-ref *propvar-enqued-value* varnr)))
    ;; (set! queue-val '()) 
    (if (null? queue-val)
        ; enque normally
        (begin
          (set! *prop-var-assignment-queue* 
                (cons (cons varnr varval) *prop-var-assignment-queue*) )
          (vector-set! *propvar-enqued-value* varnr varval) 
          #f)      
        ; either already enqued with same val or contradiction found
        (if (eq? queue-val varval)
            #f   ; do nothing
            #t))))  ; contradiction found
            

(define (prop-undo-assignments assignmentlist)  
  (let ((tmp '())
        (stopflag #f) 
        (varnr '()) ) 
   (do ((part assignmentlist (cdr part)))
       ((null? part))
     (set! tmp (car part)) 
     (set! varnr (car tmp))
     (vector-set! *propvar-value* varnr '())
     (vector-set! *propvar-enqued-value* varnr '())
     (if (cdr tmp)
        ; if var set to #t, loop over all negative occurrences
        (do ((occlist (vector-ref *propvar-negative-occ-clauses* varnr)
                      (cdr occlist) ))
            ((or (null? occlist) stopflag))
          (set-car! (car occlist) (+ (caar occlist) 1)) 
          ; the same place where we stopped cause of contradiction?
          (if #f ; (eq? *prop-unsat-found-occlistpart* occlist)
              (begin               
                (set! stopflag #t) )))     
        ; if var set to #f, loop over all positive occurrences   
        (do ((occlist (vector-ref *propvar-positive-occ-clauses* varnr)
                      (cdr occlist) ))   
            ((or (null? occlist) stopflag))
          (set-car! (cdar occlist) (+ (cadar occlist) 1)) 
          ; the same place where we stopped cause of contradiction?
          (if #f ; (eq? *prop-unsat-found-occlistpart* occlist)
              (begin 
                (set! stopflag #t) )))))))                  
   

(define (prop-undo-queue-assignments queue)
   (do ((part queue (cdr part)))
       ((null? part))
     (vector-set! *propvar-enqued-value* (caar part) '()) ))


(define (test-prop-undo-assignments assignmentlist)  
  (let ((tmp '())
        (stopflag #f) 
        (varnr '()) ) 
   ; clear set vars     
   (do ((part assignmentlist (cdr part)))
       ((null? part))
     (set! tmp (car part)) 
     (set! varnr (car tmp))
     (vector-set! *propvar-value* varnr '())
     (vector-set! *propvar-enqued-value* varnr '()) )))

        
     
   
     
;----------------------------------------------------------------------
; 
;         test stuff       
;
;----------------------------------------------------------------------

(define *prop-in*
;'(
;((1 3 4) 2)
;(() 1) 
;((1) 3 2 5)
;((1) 4 2 5)
;((2)) )) 
 
;'( 
;((1 2))
;((1) 2)
;(() 1 2)
;((2) 1) ))
    
'( 
((1 2 3))
((1  2) 3)
((1 3) 2)
((2 3) 1)
((1) 2 3)
((2) 1 3)
((3) 1 2)
(() 1 2 3) ))

    
(define *prop-internal-clauses* '())
(define *prop-maxvar* 0)
    
(define (tstprop)
  (let ((maxvar 5)
        (internalclauses '())
        (res '()) )
    
    (create-prop-var-vectors maxvar)
    
    ;(initialize-prop-var-values maxvar)
    ;(newline)(display "-- prop values initialised")(newline)
    ;(show-prop-internal-data '() maxvar)
    ;(set! internalclauses (build-prop-internal-structure *prop-in*)) 
    ;(newline)(display "-- prop internal structure built")(newline)       
    ;(show-prop-internal-data internalclauses maxvar)
    
    (set! res (solve-prop-clauses *prop-in* maxvar))
    
    (newline)(display "-- res is: ")(display res)(newline)
    (show-prop-internal-data internalclauses maxvar) ))
    

(define (show-prop-internal-data clauses maxvar)
 (let ()
  (set! maxvar *prop-atom-maxnr*)
  (newline)
  (display "showing vars 1 to ")(display maxvar)
  (display " : val, queue, pos, neg")(newline)
  (do ((var 0 (+ 1 var)))
      ((> var maxvar)) 
   (display var)(display ": " )   
   (display (vector-ref *propvar-value* var))(display ", ")
   (display (vector-ref *propvar-enqued-value* var))(display ", ")
   (newline)(display "  pos: ")
   (display (vector-ref *propvar-positive-occ-clauses* var))
   (newline)(display "  neg: ")
   (display (vector-ref *propvar-negative-occ-clauses* var))
   (newline) )
  (newline) 
  (display "showing clauses: ")(newline) 
  (for-each (lambda(cl)(display cl)(newline)) clauses)
  (newline)
  (display "*prop-var-assignment-queue* : ")(newline)
  (display "  ")(display *prop-var-assignment-queue*)(newline)
  (display "*prop-assigned-varvalpairs* : ")(newline)
  (display "  ")(display *prop-assigned-varvalpairs*)(newline)
  (display "*prop-unsat-found-occlistpart* : ")(newline)
  (display "  ")(display *prop-unsat-found-occlistpart*)(newline) ))
  

(define (tst x)    
 (set! *prop-orig-constants* '(a b))
 (set! *prop-orig-funs* '((f . 2)(g . 1)) )
 (set! *prop-orig-preds* '((p . 1)(r . 0)) )
 ; (make-prop-interpretation-lists x #t) 
 (pretty-print *prop-interp-value-pairs*)(newline)
 (pretty-print *prop-constants-mapping*)(newline)
 (pretty-print *prop-fun-mappings*)(newline) )
 ; (check-prop-interpretations x) )
 
;----------------------------------------------------------------------
;
;       random generation functions
;
;----------------------------------------------------------------------


(define random-tap 24)
(define random-size 55)

(define (random-size-int l)
  (let ((trial (string->number (make-string l #\f) 16)))
  (if (and (exact? trial) (>= most-positive-fixnum trial))
      l
      (random-size-int (- l 1)))))
(define random-chunk-size (* 4 (random-size-int 8)))

(define random-MASK
  (string->number (make-string (quotient random-chunk-size 4) #\f) 16))

(define *random-state*
  '#(
 "d909ef3e" "fd330ab3" "e33f7843" "76783fbd" "f3675fb3"
 "b54ef879" "0be45590" "a6794679" "0bcd56d3" "fabcdef8"
 "9cbd3efd" "3fd3efcd" "e064ef27" "dddecc08" "34444292"
 "85444454" "4c519210" "c0366273" "54734567" "70abcddc"
 "1bbdac53" "616c5a86" "a982efa9" "105996a0" "5f0cccba"
 "1ea055e1" "fe2acd8d" "1891c1d4" "e6690270" "6912bccc"
 "2678e141" "61222224" "907abcbb" "4ad6829b" "9cdd1404"
 "57798841" "5b892496" "871c9cd1" "d1e67bda" "8b0a3233"
 "578ef23f" "28274ef6" "823ef5ef" "845678c5" "e67890a5"
 "5890abcb" "851fa9ab" "13efa13a" "b12278d6" "daf805ab"
 "a0befc36" "0068a7b5" "e024fd90" "a7b690e2" "27f3571a"
 0))

(let ((random-strings *random-state*))
  (set! *random-state* (make-vector (+ random-size 1) 0))
  (let ((nibbles (quotient random-chunk-size 4)))
    (do ((i 0 (+ i 1)))
        ((= i random-size))
      (vector-set!
       *random-state* i
       (string->number (substring (vector-ref random-strings i)
                                  0 nibbles)
                       16)))))

;;; random-chunk returns an integer in the range of
;;; 0 to (- (expt 2 random-chunk-size) 1)
(define (random-chunk v)
  (let* ((p (vector-ref v random-size))
         (ans (logxor
               (vector-ref v (modulo (- p random-tap) random-size))
               (vector-ref v p))))
    (vector-set! v p ans)
    (vector-set! v random-size (modulo (- p 1) random-size))
    ans))

(define (random-random modu . args)
  (let ((state (if (null? args) *random-state* (car args))))
    (if #t
        (do ((ilen 0 (+ 1 ilen))
             (s random-MASK
                (+ random-MASK (* (+ 1 random-MASK) s))))
            ((>= s (- modu 1))
             (let ((slop (modulo (+ s (- 1 modu)) modu)))
               (let loop ((n ilen)
                          (r (random-chunk state)))
                 (cond ((not (zero? n))
                        (loop (+ -1 n)
                              (+ (* r (+ 1 random-MASK))
                                 (random-chunk state))))
                       ((>= r slop) (modulo r modu))
                       (else (loop ilen (random-chunk state))))))))

        1)))

(define (random-make-random-state . args)
  (let ((state (if (null? args) *random-state* (car args))))
    (list->vector (vector->list state))))



;======================= EOF resprop.scm ====================================
