
(require 'alist)
(require 'common-list-functions)
(require 'object->string)
(require 'alist-table)
(require 'relational-database)
(define build:error slib:error)

;;; This should be retrieved from site configuration database.
(define build:platform 'linux)

;;; Open the C (compiler, linker, archiver) database.
(define relational-alist-system (make-relational-system alist-table))
(define build:db ((relational-alist-system 'open-database) "ccnfigdb.scm" #f))
(define build:open-table (build:db 'open-table))
(define build:c-libraries (build:open-table 'c-libraries #f))
(define build:syscom (build:open-table 'system-commands #f))

;;; Dynamic variables.  dry-run doesn't `plan' operations, just
;;; returns the appropriate names.

(define build:dry-run #f)
(define build:using-command-line-defines #f)

;;; Look up command on a platform, but default to 'default if not
;;; initially found.

(define (make-defaulting-platform-lookup getter)
  (lambda (thing)
    (define (look platform)
      (let ((ans (getter thing platform)))
	(cond (ans ans)
	      ((eq? 'default platform)
	       (build:error "Couldn't find: " thing))
	      (else (look 'default)))))
    (look build:platform)))
(define build:lib-cc-flag
  (make-defaulting-platform-lookup (build:c-libraries 'get 'compiler-flags)))
(define build:lib-ld-flag
  (make-defaulting-platform-lookup (build:c-libraries 'get 'link-lib-flag)))

;;; plan-command is special because we want both procedure and
;;; result-name to default together, not spearately.  And because the
;;; result-name is returned, whether or not the command is planned.

(define (plan-command command)
  (define (plan platform)
    (let ((proc ((build:syscom 'get 'procedure) command platform))
	  (name ((build:syscom 'get 'result-name) command platform)))
      (cond (proc (lambda args (apply proc args) (apply name args)))
	    ((eq? 'default platform)
	     (build:error 'plan-command "unsupported command: " command))
	    (else (plan 'default)))))
  (plan build:platform))

(define (plan:flatten strings)
  (apply
   append (map
	   (lambda (obj)
	     (cond ((eq? "" obj) '())
		   ((string? obj) (list obj))
		   ((eq? #f obj) '())
		   ((null? obj) '())
		   ((list? obj) (plan:flatten obj))
		   (else (build:error 'plan:flatten "unexpected type"
				     obj "in" strings))))
	   strings)))

(define system:success? zero?)

(define (plan-to . args)
  ;;(print (plan:flatten args))
  (let ((comstr (apply string-join " " (plan:flatten args))))
    (display comstr)
    (newline)
    (cond ((not build:dry-run)
	   (let ((sts (system comstr)))
	     (cond ((system:success? sts))
		   (else (build:error 'failed: sts comstr))))))
    ))

(define (string-join joiner . args)
  (if (null? args) ""
      (apply string-append
	     (car args)
	     (map (lambda (s) (string-append joiner s)) (cdr args)))))

(define (scm:c-libs options)
  (let ((c-libs (assq 'c-lib options)))
    (if c-libs (map build:lib-ld-flag (map car (cdr c-libs))) '())))

(define (scm:includes options)
  (let ((c-libs (assq 'c-lib options)))
    (if c-libs (map build:lib-cc-flag (map car (cdr c-libs))) '())))

(define (scm:c-files options)
  (let ((c-files (assq 'c-file options)))
    (if c-files (map car (cdr c-files)) '())))

(define (scm:targets options)
  (let ((targets (assq 'target options)))
    (if targets (cdr targets) '())))

(define (scm:defines options)
  (let ((defines (assq 'define options)))
    (if defines (cdr defines) '())))

(define (scm:inits options)
  (let ((init-procs (assq 'init options)))
    (if init-procs
	(list (list "INITS" (apply string-append
				   (map (lambda (c)
					  (string-append (car c) "();"))
					(cdr init-procs)))))
	'())))

(define (command-line-defines defines)
  (if build:using-command-line-defines
      (map (lambda (def)
	     (cond ((null? (cdr def))
		    (string-append "-D" (car def)))
		   (else (string-append "-D" (car def) "="
					(object->string (cadr def))))))
	   defines)
      ""))
(define (change-suffix str old new)
  (let* ((len (string-length str))
	 (re (- len (string-length old))))
    (cond ((string-ci=? old (substring str re len))
	   (string-append (substring str 0 re) new))
	  (else
	   (build:error 'change-suffix "suffix doens't match:" old str)))))

(define (write-c-defines-to-file defines file)
(if (and (not build:using-command-line-defines)
	 (not build:dry-run))
    (call-with-output-file file
      (lambda (p)
	(newline p)
	(for-each (lambda (d)
		    (display (apply string-join " " "#define" d) p)
		    (newline p))
		  defines)))))

;;;	Example ways scm:build might be called:
;;; (scm:build '((executable scm) (feature curses) (feature dynamic-linking)
;;;		 (object "sc2") (feature inexact) (feature bignums)))
;;; (scm:build '((dlll curses)))

(define (scm:build specs)
  (define options (scm:specs->options specs))
  (let ((targets (scm:targets options))
	(includes (scm:includes options))
	(defines (append (scm:inits options) (scm:defines options)))
	(huhs (remove-if
	       (lambda (p) (memq (car p) '(define target c-file c-lib init)))
	       options)))
    (if (not (null? huhs))
	(build:error 'scm:build "unknown attributes: " huhs))
    (map
     (lambda (target)
       (case (car target)
	 ((executable)
	  (let ((link-c-program (plan-command 'link-c-program))
		(compile-c-file (plan-command 'compile-c-file)))
	    (write-c-defines-to-file defines "scmflags.h")
	    (link-c-program
	     (cadr target)
	     (map (lambda (f) (compile-c-file f includes defines))
		  (scm:c-files options))
	     (scm:c-libs options))))
	 ((library)
	  (let ((make-archive (plan-command 'make-archive))
		(compile-c-file (plan-command 'compile-c-file)))
	    (write-c-defines-to-file defines "scmflags.h")
	    (make-archive
	     (cadr target)
	     (map (lambda (f) (compile-c-file f includes defines))
		  (scm:c-files options)))))
	 ((dlll)
	  (let ((make-dll-archive (plan-command 'make-dll-archive))
		(compile-dll-c-file (plan-command 'compile-dll-c-file)))
	    (write-c-defines-to-file defines "scmflags.h")
	    (make-dll-archive
	     (cadr target)
	     (map (lambda (f) (compile-dll-c-file f includes defines))
		  (scm:c-files options)))))))
     targets)))

(define (build:group-by-attribute specs)
  (define al '())
  (define add-it
    (let ((getter (alist-inquirer eq?))
	  (putter (alist-associator eq?)))
      (lambda (key val)
	(set! al (putter al key (adjoin val (or (getter al key) '())))))))
  (for-each (lambda (spec)
	      (add-it (car spec) (cdr spec)))
	    specs)
  al)

(define (scm:specs->options specs)
  (build:group-by-attribute
   (apply append (map scm:expand-spec specs))))

(define (scm:expand-spec spec)
  (let ((type (car spec))
	(item (cadr spec)))
    (define (huh?)
      (build:error 'scm:expand-spec "don't understand: " type item))
    (case type
      ((executable)
       (case item
	 ((scm)
	  (cons '(target executable "scm")
		(cons (list 'define "IMPLINIT"
			    (object->string
			     (in-vicinity
			      (implementation-vicinity) "Init.scm")))
		      (map (lambda (f) (list 'c-file (string-append f ".c")))
			   '("eval" "repl" "scl" "scm" "subr"
				    "sys" "time" "unif")))))
	 (else (huh?))))
      ((library)
       (case item
	 ((scm) '(target library "scm") '((define "RTL")))
	 (else (huh?))))
      ((dlll)
       (case item
	 ((curses)
	  '((target dlll "crs") (c-file "crs.c") (init "init_crs")))
	 ((i/o-extensions)
	  '((target dlll "ioext") (c-file "ioext.c") (init "init_ioext")))
	 ((posix)
	  '((target dlll "posix") (c-file "posix.c") (init "init_posix")))
	 ((record)
	  '((target dlll "record") (c-file "record.c") (init "init_record")))
	 ((regex)
	  '((target dlll "rgx") (c-file "rgx.c") (init "init_rgx")))
	 ((rev2-procedures)
	  '((target dlll "sc2") (c-file "sc2.c") (init "init_sc2")))
	 ((socket)
	  '((target dlll "socket") (c-file "socket.c") (init "init_socket")))
	 ((turtlegr)
	  '((target dlll "turtlegr")
	    (c-file "turtlegr.c") (init "init_turtlegr")))
	 ((unix)
	  '((target dlll "unix") (c-file "unix.c") (init "init_unix")))
	 ((array-map)
	  '((target dlll "ramap") (c-file "ramap.c") (init "init_ramap")
				  (define "ARRAYS")))
	 ((generalized-c-arguments)
	  '((target dlll "gsubr") (c-file "gsubr.c") (init "init_gsubr")))
	 (else (huh?))))
      ((feature)
       (case item
	 ((reckless) '((define "RECKLESS")))
	 ((cautious) '((define "CAUTIOUS")))
	 ((careful-interrupt-masking) '((define "CAREFUL_INTS")))
	 ((bignums) '((define "BIGNUMS")))
	 ((arrays) '((define "ARRAYS")))
	 ((array-map) '((c-file "ramap.c") (init "init_ramap")
					   (define "ARRAYS")))
	 ((inexact) '((define "FLOATS") (c-lib m)))
	 ((engineering-notation) '((define "ENGNOT")))
	 ((single-precision-only) '((define "SINGLESONLY")))
	 ((compiled-closure) '((define "CCLO")))
	 ((sicp) '((define "SICP")))
;;;	 ((initial-heap-size) '((define "STACK_LIMIT" (* 25000 sizeof-cell))))
;;;	 ((heap-segment-size) '((define "HEAP_SEG_SIZE" (* 8100 sizeof-cell))))
;;;	 ((short-aligned-stack) '((define "SHORT_ALIGN")))
;;;	 ((initial-malloc-limit) '((define "INIT_MALLOC_LIMIT" 100000)))
;;;	 ((number-of-hash-buckets) '((define "NUM_HASH_BUCKETS" 137)))
;;;	 ((minimum-gc-yield) '((define "MIN_GC_YIELD" "(heap_size/4)")))
	 ((stack-limit) '((define "STACK_LIMIT" "(HEAP_SEG_SIZE/2)")))
	 ((heap-can-shrink) '((define "GC_FREE_SEGMENTS")))
	 ((cheap_continuations) '((define "CHEAP_CONTINUATIONS")))
	 ((memoize-local-bindings) '((define "MEMOIZE_LOCALS")))
	 ((tick-interrupts) '((define "TICKS")))
	 ((poll) '((define "NON_PREEMPTIVE")))

	 ((windows) '((c-lib windows) (define "PROT386")))
	 ((dynamic-linking)
	  '((c-file "dynl.c") (c-lib dlll) (init "init_dynl")))

	 ((curses) '((c-file "crs.c") (c-lib curses) (init "init_crs")))
	 ((i/o-extensions) '((c-file "ioext.c") (init "init_ioext")))
	 ((posix) '((c-file "posix.c") (init "init_posix")))
	 ((record) '((c-file "record.c") (init "init_record")))
	 ((regex) '((c-file "rgx.c") (c-lib regex) (init "init_rgx")))
	 ((rev2-procedures) '((c-file "sc2.c") (init "init_sc2")))
	 ((socket) '((c-file "socket.c") (init "init_socket")))
	 ((turtlegr)
	  '((c-file "turtlegr.c") (c-lib graphics) (init "init_turtlegr")))
	 ((unix) '((c-file "unix.c") (init "init_unix")))
	 ((generalized-c-arguments) '((c-file "gsubr.c") (init "init_gsubr")))
	 (else (huh?))))
      ((object) (list 'c-file item))
      (else (huh?)))))

;(write (scm:build '((executable scm)
;		    (feature dynamic-linking) (feature inexact)
;		    (feature bignums) (feature arrays))))
;(write (scm:build '((dlll socket))))
;(newline)
